;;;; File: er-macros.scm
;;;; Author: Juergen Lorenz
;;;; ju (at) jugilo (dot) de
;;;; Date: Oct 24, 2010
;;;;       Oct 28, 2010
;;;;       Nov 02, 2010
;;;;       Nov 04, 2010
;;;;       Nov 09, 2010
;;;;       Nov 26, 2010
;;;;       Nov 28, 2010
;;;;       Jan 19, 2011
;;;;       Jan 27, 2011
;;;;       Feb 02, 2011
;;;;       Feb 16, 2011
;;;;       Feb 18, 2011
;;;;       Jun 20, 2011
;;;;       Jun 22, 2011
;;;;       Jun 26, 2011
;;;;       Jul 13, 2011


;The macros in this module have been renamed for consistency with the
;ir-macros module. The old names are deprecated, but still implemented
;as aliases, so that code relying on them is not broken.
;In fact, the ir-macros module is much more convenient.

(module er-macros *
(import scheme (only chicken gensym assert condition-case))
(import-for-syntax (only matchable match))

;As an application of the match macro we'll define a macro,
;er-macro-rules, that will simplify the implementation of low-level
;macros considerably. Note, that the transformer of such a macro is a
;routine of three parameters, usually called form, rename, compare?, and
;that within that routine form has to be destructured by hand, and the
;symbols have to be renamed by hand as well. In our enhancement, we want
;both to be done (almost) automatically. Moreover, we want our macro to
;be of the same syntax as syntax-rules, i.e. it should look like this
;
;  (er-macro-rules (?%sym ...) (pat xpr) (pat1 xpr1) ...)
;
;where %sym explicitly renames the symbol sym (another one letter prefix
;would do as well), pat, pat1, ... represent admissible macro calls and
;xpr, xpr1 ... corresponding expressions, evaluationg to the
;macro-expansion.  Moreover, er-macro-rules
;should expand into a (lambda (form rename compare?) ...) as does
;syntax-rules. Since match consumes the form argument and %sym ... the
;rename argument, only the remaining compare? needs to be available to
;the macro-writer. Our trick is to make all er-macro-... macros in this
;module unhygienic. They all pollute their local namespace with the
;compare? symbol.

;Note that the list (%sym ...) is in general rather long, it will be
;transformed into a long let of the form (let ((%sym (rename (sym-cdr
;'%sym))) ...) in the expansion, where sym-cdr is a local routine, which
;strips the prefix.
;
;As a non-trivial example consider the following low-level
;implementation of or. Note the similarity to the high-level one.
;
;  (define-syntax my-or
;    (er-macro-rules (%if %my-or)
;      ((_) #f)
;      ((_ arg . args)
;       `(,%if ,arg ,arg (,%my-or ,@args)))))

;Since this macro is unhygienic, it must be implemented as low-level
;macro.
;
;;; (er-macro-rules (%sym ...) (pat0 xpr0) (pat1 xpr1) ...)
;;; -------------------------------------------------------
;;; where xpr0, xpr1, ... are expressions which generate the macro
;;; expansion.
;;; Checks the macro's use against a series of patterns, pat0, pat1 ...
;;; and returns the corresponding macro-expansion.
(define-syntax er-macro-rules
  (er-macro-transformer
  (lambda (f r c?)
    (let (
      (syms (cadr f))
      (rules (cddr f))
      )
      (let (
        (pats (map car rules))
        (xprs (map cadr rules))
        (sym-cdr
          (lambda (%sym)
            (string->symbol 
              (substring
                (symbol->string %sym) 1))))
        (%let (r 'let))
        (%lambda (r 'lambda))
        (%match (r 'match))
        (%er-macro-transformer (r 'er-macro-transformer))
        (gform (gensym 'form))
        (grename (gensym 'rename))
        )
        `(,%er-macro-transformer
           (,%lambda (,gform ,grename compare?)
             (,%let ,(map (lambda (sym)
                            `(,sym (,grename ',(sym-cdr sym))))
                          syms)
               (,%match ,gform
                 ,@(map (lambda (pat xpr) `(,pat ,xpr))
                        pats xprs))))))))))

;With er-macro-rules under our belt, it's easy to implement three
;macros, er-macro-define, er-macro-let and er-macro-letrec, which
;facilitate the implementation of low-level macros even more: We simply
;match one pattern, the macro code, against a list of the form
;  (with (%sym ...) . body)
;This time, my-or would look like this
;
;  (er-macro-define (my-or . args)
;    (with (%if %my-or)
;      (if (null? args)
;        #f
;        (let ((tmp (car args)))
;          `(,%if ,tmp ,tmp (,%my-or ,@(cdr args)))))))

;;; (er-macro-define code (with (sym% ...) . body))
;;; -----------------------------------------------
;;; where code is the complete macro-code (name . args), i.e. the pattern
;;; of a macro call, %sym ... are aliases of sym ... and body a list of
;;; expressions xpr . xprs which produce the macro-expansion.

(define-syntax er-macro-define
  (syntax-rules (with)
    ((_ (name . args) (with syms xpr . xprs))
     (define-syntax name
       (er-macro-rules syms
         ((_ . args) (begin xpr . xprs)))))))

;er-macro-let and er-macro-letrec are local versions of er-macro-define, where
;the local macros are evaluated in parallel or recursively. For example

;  (let ((f (lambda (n) (+ n 10))))
;    (er-macro-let (
;      ((f n) n)
;      ((g n) `(,%f ,n))
;      )
;      (display (list (f 1) (g 1))) (newline)))
;
;will result in (1 11) while  
;
;  (let ((f (lambda (n) (+ n 10))))
;    (er-macro-letrec (
;      ((f n) n)
;      ((g n) `(,%f ,n))
;      )
;      (display (list (f 1) (g 1))) (newline)))
;
;returns (1 1).

;;; (er-macro-let ((code0 (with (%sym0 ...) . body0) ...) . body)
;;; -------------------------------------------------------------
;;; where code0, %sym0 and body0 are as in er-macro-define. This is
;;; a local version of er-macro-define, allowing a list of
;;; (code with-xpr) lists to be processed in body in parallel.
(define-syntax er-macro-let
  (syntax-rules (with)
    ((_ (
       ((name0 . args0) (with syms0 xpr0 . xprs0))
       ...
       )
       xpr . xprs)
     (let-syntax (
       (name0
         (er-macro-rules syms0
           ((_ . args0) (begin xpr0 . xprs0))))
       ...
       ) xpr . xprs))))


;;; (er-macro-letrec ((code0 (with (%sym0 ...) . body0) ...) . body)
;;; -------------------------------------------------------------
;;; where code0, %sym0 and body0 are as in er-macro-define. This is
;;; a local version of er-macro-define, allowing a list of
;;; (code with-xpr) lists to be processed in body recursively.
(define-syntax er-macro-letrec
  (syntax-rules (with)
    ((_ (
       ((name0 . args0) (with syms0 xpr0 . xprs0))
       ...
       )
       xpr . xprs)
     (letrec-syntax (
       (name0
         (er-macro-rules syms0
           ((_ . args0) (begin xpr0 . xprs0))))
       ...
       ) xpr . xprs))))


;The following macro can be used to simplify renaming.
;Its syntax is similar to Graham's with-gensyms but more flexible
;because of the extra rename argument.
;
;;; (with-aliases (rename %sym0 %sym1 ...) . body)
;;; ----------------------------------------------
;;; executes body with aliases of sym0 sym1 ...
;;; (to be used in macros with import-for-syntax)
(define-syntax with-aliases
  (er-macro-transformer
    (lambda (f r c?)
      (let (
        (rename (caadr f))
        (%syms (cdadr f))
        (body (cddr f))
        (%let (r 'let))
        )
        (let ((syms (map (lambda (%sym)
                            (string->symbol
                              (substring (symbol->string %sym) 1)))
                    %syms)))
          `(,%let ,(map (lambda (%sym sym) `(,%sym (,rename ',sym)))
                      %syms syms)
             ,@body))))))

) ; module er-macros

